#include <tigcclib.h>

#define RANDOM_STRENGTH 4
#define DATA_NAME SYMSTR("enigdata")

#include "extrand.h"

typedef struct
{
	char initials[3];
	unsigned char random_strength;
} FILE_VERSION;

typedef struct
{
	struct
	{
		unsigned short size;
		FILE_VERSION version;
	} header;
	struct
	{
		RANDOM_BUFFER buffer;
	} data;
	struct
	{
		char type[7];
	} footer;
} DATA_FILE;

/*function prototypes*/
static char Adjust(char, char);
static char Wrap(char);
static char Scramble(char *, char);

void _main(void)
{
	unsigned char *input_string, *output_string = NULL;
	unsigned short i, temp; /*a counter for the loops and temporary storage*/
	unsigned long seed1, seed2; /*the random seeds*/
	DATA_FILE *filep = NULL;
	ESI argptr;
	FILE_VERSION Check = {{'P', 'J', 'R'}, RANDOM_STRENGTH};
	SYM_ENTRY *file_vat;

	file_vat = SymFindPtr(DATA_NAME, 0); /*try to find the data file that stores the
	state of the random number generator*/
	if(file_vat != NULL) /*if the file exists*/
	{
		if(file_vat->flags.bits.archived) /*and it is archived*/
		{
			EM_moveSymFromExtMem(DATA_NAME, HS_NULL); /*unarchive it*/
		}
		filep = HeapDeref(file_vat->handle);
	}
	if(file_vat == NULL || memcmp(HeapDeref(file_vat->handle) + 2, &Check,
	sizeof(FILE_VERSION))) /*if it does not exist or is the wrong version*/
	{
		RANDOM_BUFFER buffer;

		randomize_long();

		SaveRandState(&buffer);

		/*create it*/
		SymDel(DATA_NAME);
		file_vat = DerefSym(SymAdd(DATA_NAME));
		file_vat->handle = HeapAlloc(sizeof(DATA_FILE));
		filep = HeapDeref(file_vat->handle);

		filep->header.size = sizeof(DATA_FILE) - 2;
		memcpy(&filep->header.version, &Check, sizeof(FILE_VERSION));
		memcpy(&filep->data.buffer, &buffer, sizeof(RANDOM_BUFFER));
		memcpy(filep->footer.type, &(unsigned char[]){0, 'D', 'a', 't', 'a', 0, OTH_TAG}, 7);
	}

	LoadRandState(&filep->data.buffer);

	InitArgPtr(argptr); /*sets the argptr to point to the first argument of the program*/

	if(GetArgType(argptr) == STR_TAG) /*if the first argument is a string*/
	{
		input_string = (unsigned char *)GetStrnArg(argptr); /*get the string*/
		if(input_string[0] == 0xA9) /*if the string needs to be decoded*/
		{
			if(strlen(input_string) <= 7) /*if the string was not encoded by this
			program*/
			{
				while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the
				program off the stack*/
				{
					top_estack = next_expression_index(top_estack);
				}
				top_estack--;

				push_expr_quantum(SYMSTR("Not an Enigma coded string"), STR_TAG);
				/*push_ANSI_string only returns one word.  Thanks to KronicDeth for
				telling me how to do this*/
				return;
			}
			output_string = malloc(strlen(input_string) - 5); /*allocate memory for the
			output string*/
			if(output_string == NULL) /*if there was not enough memory*/
			{
				while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the
				program off the stack*/
				{
					top_estack = next_expression_index(top_estack);
				}
				top_estack--;

				push_expr_quantum(SYMSTR("Memory low"), STR_TAG);
				return;
			}
			output_string[0] = 0; /*put a null character at the beginning of the output
			string*/

			seed1 = 2704L * Adjust(input_string[1], 0) + 52L * Adjust(input_string[2], 0)
			+ Adjust(input_string[3], 0); /*convert the number represented by the first
			three characters from base fifty-two to base ten*/
			seed2 = 2704L * Adjust(input_string[4], 0) + 52L * Adjust(input_string[5], 0)
			+ Adjust(input_string[6], 0); /*convert the number represented by the second
			three characters from base fifty-two to base ten*/
			input_string += 7; /*get rid of the first seven characters of the input string
			(the comment symbol and the seeds)*/

			srand_long(seed1); /*seed the random function with the number represented by
			the first three (excluding the comment symbol) characters of the string so
			that it will generate the same random sequence again.*/

			for(i = 0; i < strlen(input_string); i++)
			{
				if((input_string[i] >= 65 && input_string[i] <= 90) || (input_string[i] >=
				97 && input_string[i] <= 122)) /*if the current character is a letter*/
				{
					temp = Wrap(Adjust(input_string[i], 0) - random_long(52));
					/*undo the random scrambling*/
					input_string[i] = Adjust(temp, 1);
				}
			}

			srand_long(seed2); /*seed the random function with the number represented by
			the first three (excluding the comment symbol) characters of the string so
			that it will generate the same random sequence again.*/

			temp = Scramble(input_string, 1);
			if(temp == 1)
			{
				free(output_string);

				while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the
				program off the stack*/
				{
					top_estack = next_expression_index(top_estack);
				}
				top_estack--;

				push_expr_quantum(SYMSTR("Memory low"), STR_TAG);
				return;
			}

			memcpy(output_string + 1, input_string, strlen(input_string) + 1); /*copy the
			input string, including the null character, to the output string*/

			while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the program
			off the stack*/
			{
				top_estack = next_expression_index(top_estack);
			}
			top_estack--;

			push_expr_quantum(output_string + strlen(output_string + 1) + 1, STR_TAG);
			/*returns the output string*/
		}
		else /*the input string needs to be encoded*/
		{
			output_string = malloc(strlen(input_string) + 9); /*allocate memory for the
			output string*/
			if(output_string == NULL) /*if there was not enough memory*/
			{
				while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the
				program off the stack*/
				{
					top_estack = next_expression_index(top_estack);
				}
				top_estack--;

				push_expr_quantum(SYMSTR("Memory low"), STR_TAG);
				return;
			}
			output_string[0] = 0;
			/*put a null character and a comment symbol at the beginning of the string*/
			output_string[1] = 0xA9;

			srand_long(seed2 = random_long(140607)); /*seed the random function with a
			random number and remember what the number was*/

			temp = Scramble(input_string, 0);
			if(temp == 1)
			{
				free(output_string);

				while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the
				program off the stack*/
				{
					top_estack = next_expression_index(top_estack);
				}
				top_estack--;

				push_expr_quantum(SYMSTR("Memory low"), STR_TAG);
				return;
			}

			srand_long(seed1 = random_long(140607)); /*seed the random function with a
			random number and remember what the number was*/

			/*convert the seeds to base 52 and put them in the string*/
			output_string[2] = Adjust(seed1 / 2704, 1);
			seed1 -= seed1 / 2704 * 2704;
			output_string[3] = Adjust(seed1 / 52, 1);
			seed1 -= seed1 / 52 * 52;
			output_string[4] = Adjust(seed1, 1);
			output_string[5] = Adjust(seed2 / 2704, 1);
			seed2 -= seed2 / 2704 * 2704;
			output_string[6] = Adjust(seed2 / 52, 1);
			seed2 -= seed2 / 52 * 52;
			output_string[7] = Adjust(seed2, 1);

			for(i = 0; i < strlen(input_string); i++)
			{
				if((input_string[i] >= 65 && input_string[i] <= 90) || (input_string[i] >=
				97 && input_string[i] <= 122)) /*if the current character is a letter*/
				{
					temp = Wrap(Adjust(input_string[i], 0) + random_long(52));
					/*randomly scramble the letters*/
					input_string[i] = Adjust(temp, 1);
				}
			}
			memcpy(output_string + 8, input_string, strlen(input_string) + 1); /*copy the
			input string, including the null character, to the output string*/

			while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the program
			off the stack*/
			{
				top_estack = next_expression_index(top_estack);
			}
			top_estack--;

			push_expr_quantum(output_string + strlen(output_string + 1) + 1, STR_TAG);
			/*return the output string*/
		}
	}
	else /*if there were no arguemnts or the first argument was not a string*/
	{
		while(GetArgType(top_estack) != END_TAG) /*cleans the arguments of the program off
		the stack*/
		{
			top_estack = next_expression_index(top_estack);
		}
		top_estack--;

		push_expr_quantum(SYMSTR("Argument must be a string"), STR_TAG);
	}
	if(output_string != NULL) /*free the memory for the output string*/
	{
		free(output_string);
	}

	SaveRandState(&filep->data.buffer);
	EM_moveSymToExtMem(DATA_NAME, HS_NULL); /*archive the data file*/
}

char Adjust(char letter, char mode) /*either converts a letter to is numeric value or
coverts a number to its corresponding letter*/
{
	if(mode == 0)
	{
		return(letter - (letter <= 90 ? 65 : 71));
	}
	else
	{
		return(letter + (letter <= 25 ? 65 : 71));
	}
}

char Wrap(char input) /*wraps the letters around*/
{
	return(input >= 0 ? (input <= 51 ? input : input - 52) : input + 52);
}

char Scramble(char *input_string, char mode)
{
	char *output_string;
	unsigned short i, temp, length;

	length = strlen(input_string);

	output_string = malloc(length + 1);
	if(output_string == NULL)
	{
		return(1);
	}
	memset(output_string, 0, length + 1);

	for(i = 0; i < length; i++)
	{
		if(mode == 0)
		{
			do
			{
				temp = random_long(length);
			} while(output_string[temp] != 0); /*find an empty spot in the output
			string*/
			output_string[temp] = input_string[i]; /*put the next character there*/
		}
		else
		{
			do
			{
				temp = random_long(length);
			} while(input_string[temp] == 0);
			output_string[i] = input_string[temp];
			input_string[temp] = 0;
		}
	}

	memcpy(input_string, output_string, length + 1);
	return(0);
}
